//----------------------------------------------------------------------------
//
// File:	convert.cpp
// Date:      	09-May-1995
// Programmer:	Marc Rousseau
//
// Description: This programs will convert a .lst or .dat file to a .ctg file
//
// Copyright (c) 1994-2000 Marc Rousseau, All Rights Reserved.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
//
// Revision History:
//
//----------------------------------------------------------------------------

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "common.hpp"
#include "cartridge.hpp"

#if defined ( __GNUC__ )
  extern void swab ( char *, char *, unsigned );
#endif

enum eFileType {
    LST,
    ROM,
    HEX,
    GK,
    CTG
};

struct sGromHeader {
    // This is only valid AFTER bytes have been swapped !!!
    UCHAR    Valid;
    UCHAR    Version;
    UCHAR    NoApplications;
    UCHAR    reserved1;
    USHORT   PowerUpHeader;
    USHORT   ApplicationHeader;
    USHORT   DSR_Header;
    USHORT   SubprogramHeader;
    USHORT   InterruptHeader;
    USHORT   reserved2;

    sGromHeader ( UCHAR *ptr );
};

sGromHeader::sGromHeader ( UCHAR *ptr )
{
    sGromHeader *rawHdr = ( sGromHeader * ) ptr;
    memcpy (( char * ) this, ptr, sizeof ( *this ));
    swab (( char * ) &rawHdr->PowerUpHeader, ( char * ) &this->PowerUpHeader, 6 * sizeof ( USHORT ));
}

void FindName ( cCartridge &cart )
{
    if ( cart.m_Title != NULL ) return;

    printf ( "Found the following names:\n" );

    char *title = NULL;
    for ( unsigned i = 0; i < SIZE ( cart.GromMemory ); i++ ) {
        sMemoryRegion *memory = &cart.GromMemory[i];
        if ( memory->noBytes[0] == 0 ) continue;
        sGromHeader hdr ( memory->Bank[0] );
        if ( hdr.Valid != 0xAA ) continue;
        int addr = i * 0x2000;
        if ( hdr.ApplicationHeader == 0 ) continue;
        int appHdr = hdr.ApplicationHeader;
        while ( appHdr ) {
            appHdr -= addr;
            UCHAR length = memory->Bank[0][ appHdr + 4 ];
            UCHAR *name = &memory->Bank[0][ appHdr + 5 ];
            printf ( " %c %*.*s\n", title ? ' ' : '\x10', length, length, name );
            appHdr = ( memory->Bank[0][ appHdr ] << 8 ) + memory->Bank[0][ appHdr + 1];
            if ( title == NULL ) {
                title = ( char * ) malloc ( length + 1 );
                sprintf ( title, "%*.*s", length, length, name );
            }
        }
    }
    cart.m_Title = title;

    printf ( "\n" );
}

// See if we can cut GROMs from 8k to 6k ( or less )
void CheckGrom ( cCartridge &cart )
{
    for ( unsigned i = 0; i < SIZE ( cart.GromMemory ); i++ ) {
        sMemoryRegion *memory = &cart.GromMemory[i];
        if ( memory->noBytes[0] == 0 ) continue;
        for ( USHORT b = 0; b < memory->noBanks; b++ ) {
            // Try to see if it's really 6K instead of 8K
            if ( memory->noBytes[b] == 0x2000 ) {
                UCHAR *ptr = memory->Bank [b] + 0x1800;
                int check = 0;
                while ( check < 0x1800 ) {
                    if ( memcmp ( memory->Bank[b] + check, ptr, 0x0800 ) == 0 ) break;
                    check += 0x0800;
                }
                if ( check != 0x1800 ) memory->noBytes[b] -= 0x0800;
            }
            // Don't save empty bytes, since we default to 00 anyway
            UCHAR *ptr = &memory->Bank [b][ memory->noBytes[b] - 1 ];
            int extra = 0;
            while ( *ptr == '\0' ) {
                extra++;
                ptr--;
            }
            memory->noBytes[b] -= extra;
        }
    }
}

//----------------------------------------------------------------------------
//
// Create a module from a .lst/.dat listing file
//
// The types of sections that can be specified are:
//
//    RAM   - Read/Write CPU memory - any initialization data specified is ignored
//    ROM   - Read-Only CPU memory
//    RAMB  - Battery backed Read/Write CPU memory
//    GRAM  - Read/Write GROM memory - any initialization data specified is ignored
//    GROM  - Read-Only GROM memory
//    GRAMB - Battery backed Read/Write GROM memory
//
// ROM/RAM banks are 4096 (0x1000) bytes
// GROM/GRAM banks are 8192 (0x2000) bytes
//
//----------------------------------------------------------------------------
char *types[6] = { "RAM", "ROM", "Backed RAM", "GRAM", "GROM", "Backed GRAM" };

int ConvertDigit ( char digit )
{
    int number = digit - '0';
    if ( number > 9 ) number -= ( 'A' - '0' ) - 10;
    return number;
}

char *GetAddress ( char *ptr, ULONG *address )
{
    *address = 0;

    if ( ptr == NULL ) return NULL;

    while ( isxdigit ( *ptr )) {
        *address <<= 4;
        *address |= ConvertDigit ( *ptr++ );
    }

    while (( *ptr != '\0' ) && ! isspace ( *ptr )) ptr++;

    return ( *ptr != '\0' ) ? ptr : NULL;
}

void ReadFile ( char *filename, cCartridge &cartridge )
{
    FILE *file = fopen ( filename, "rt" );
    if ( file == NULL ) {
        fprintf ( stderr, "Unable to open file %s\n", filename );
        exit ( -1 );
    }

    unsigned max = ( unsigned ) -1;
    for ( EVER ) {
        char line [ 256 ], type [10], temp[10];
        int base, region, banks;

        char ch = fgetc ( file );
        if ( feof ( file )) break;
        ungetc ( ch, file );

        fgets ( line, 256, file );
    
        switch ( line[0] ) {
            case ';' :
                break;
            case '[' :
                if ( cartridge.m_Title ) free ( cartridge.m_Title );
                line [ strlen ( line ) - 2 ] = '\0';
                cartridge.m_Title = strdup ( &line[1] );
            case '*' :
                continue;
            default :
                fprintf ( stderr, "File error\n" );
                continue;
        }
        sscanf ( &line[2], "%s %d", type, &region );

        fgets ( line, 256, file );
        if ( line[0] != ';' ) {
            fprintf ( stderr, "File error\n" );
            continue;
        }

        if ( strnicmp ( &line [2], "CRU", 3 ) == 0 ) {
            sscanf ( &line[2], "%s = %x", temp, &base );
            fgets ( line, 256, file );
            if ( line[0] != ';' ) {
                fprintf ( stderr, "File error\n" );
                continue;
            }
        } else {
            base = 0;
        }

        sscanf ( &line[2], "%s = %d", temp, &banks );

        char *ptr = type;
        char t = ( *ptr == 'G' ) ? 3 : 0;
        sMemoryRegion &memory = ( t ) ? cartridge.GromMemory[region] : cartridge.CpuMemory[region];
        int size = ( t ) ? 0x2000 : 0x1000;
        int mask = ( t ) ? 0x1FFF : 0x0FFF;
        if ( t ) ptr++;
        memory.Type = ( ptr[3] == 'B' ) ? _BACKED_RAM_ : 
                      ( ptr[1] == 'O' ) ? _ROM_ : _RAM_;
        memory.cruBase = base;
        memory.noBanks = banks;

        printf ( "\nReading %-4s %2d :     ", types[t+memory.Type], region );
        for ( int i = 0; i < banks; i++ ) {
            memory.noBytes[i] = ( size + 0x0800 ) & 0xF000;
            memory.Bank[i] = new UCHAR [ memory.noBytes[i] ];
            if ( memory.Type == _RAM_ ) continue;
            memset ( memory.Bank[i], 0, memory.noBytes[i] );
            memory.curBank = memory.Bank[i];
            for ( int j = 0; j < size; ) {
                fgets ( line, 256, file );
                if ( line [0] == '*' ) break;
                ULONG address;
                char *src = GetAddress ( line, &address );
                if ( src == NULL ) break;
                printf ( "%04X", address );
                UCHAR *dest = &memory.curBank [ address & mask ];
                int x = 0;
                while (( *src != '\0' ) && (( unsigned ) x < max )) {
                    while (( *src == ' ' ) || ( *src == '-' )) src++;
                    if ( isxdigit ( src [0] ) && isxdigit ( src [1] )) {
                        char high = ConvertDigit ( *src++ );
                        char low  = ConvertDigit ( *src++ );
                        if ( isxdigit ( *src )) break;
                        *dest++ = ( high << 4 ) | low;
                        if ( ++j == size ) break;
                        x++;
                    } else {
                        src++;
                    }
                }
                if (( max == ( unsigned ) -1 ) && ( x > 0 )) max = x;
            }
        }
    }
    fclose ( file );
    printf ( "\n" );
}

//----------------------------------------------------------------------------
//
// Read V9T9 style .hex/.bin files given a 'base' file name.
//
// A search is done first for V6.0 style names using the specified extension
//   i.e. Given foo.bin -> foog.bin, fooc.bin, food.bin
// If no match is found, a search for earlier naming conventions is made
//   i.e. Given foo.hex -> foo(g).hex, fooc0.hex, fooc1.hex
//
//----------------------------------------------------------------------------
void ReadHex ( char *fileName, cCartridge &cartridge )
{
    char name [ 512 ], ext [ 4 ];

    // Truncate the filename to the 'base' filename
    int len = strlen ( fileName );
    strcpy ( ext, &fileName [len-3]);
    fileName [len-4] = '\0';

    // Read GROM
    sprintf ( name, "%sg.%s", fileName, ext );
    FILE *file = fopen ( name, "rb" );
    if ( file == NULL ) {
        sprintf ( name, "%s.%s", fileName, ext );
        file = fopen ( name, "rb" );
        if ( file == NULL ) {
            fprintf ( stderr, "Unable to open file %s.\n", name );
            exit ( -1 );
        }
    }

    int region = 3;
    int ch = getc ( file );
    while ( ! feof ( file )) {
        ungetc ( ch, file );
        sMemoryRegion &memory = cartridge.GromMemory[region++];
        memory.Type = _ROM_;
        memory.noBanks = 1;
        memory.noBytes[0] = 0x2000;
        memory.Bank[0] = new UCHAR [ memory.noBytes[0] ];
        fread ( memory.Bank[0], 1, 0x2000, file );
        ch = getc ( file );
    }
    fclose ( file );

    // Read RAM 0
    sprintf ( name, "%sc.%s", fileName, ext );
    file = fopen ( name, "rb" );
    if ( file == NULL ) {
        sprintf ( name, "%sc0.%s", fileName, ext );
        file = fopen ( name, "rb" );
        if ( file == NULL ) return;
    }

    for ( int i = 6; i < 8; i++ ) {
        sMemoryRegion &memory = cartridge.CpuMemory[i];
        memory.Type = _ROM_;
        memory.noBanks = 1;
        memory.noBytes[0] = 0x1000;
        memory.Bank[0] = new UCHAR [ memory.noBytes[0] ];
        fread ( memory.Bank[0], 1, 0x1000, file );
    }
    fclose ( file );

    // Read RAM 1
    sprintf ( name, "%sd.%s", fileName, ext );
    file = fopen ( name, "rb" );
    if ( file == NULL ) {
        sprintf ( name, "%sc1.%s", fileName, ext );
        file = fopen ( name, "rb" );
        if ( file == NULL ) return;
    }

    for ( int i = 6; i < 8; i++ ) {
        sMemoryRegion &memory = cartridge.CpuMemory[i];
        memory.Type = _ROM_;
        memory.noBanks = 2;
        memory.noBytes[1] = 0x1000;
        memory.Bank[1] = new UCHAR [ memory.noBytes[1] ];
        fread ( memory.Bank[1], 1, 0x1000, file );
    }
    fclose ( file );
}

//----------------------------------------------------------------------------
//  Read V9T9 style CPU & GPL console ROMS and make a 'special' cartridge
//----------------------------------------------------------------------------
void ReadConsole ( char *fileName, cCartridge &cartridge )
{
    char name [ 512 ], *ext, *base = "";

    // Truncate the filename to the 'base' filename
    int len = strlen ( fileName );
    ext = &fileName [len-3];

    if (( len >= 6 ) && ( strnicmp ( fileName + len - 6, "ti.", 3 ) == 0 )) {
        if ( len > 6 ) {
            base = fileName;
            fileName [ len - 6 ] = '\0';
	}
        sprintf ( name, "%sticpu.%s", base, ext );
        ReadHex ( name, cartridge );
        // Move GROM 3 to ROMs 0+1
        sMemoryRegion &memory = cartridge.GromMemory[3];
        cartridge.CpuMemory[0] = memory;
        cartridge.CpuMemory[0].noBytes[0] = 0x1000;
        cartridge.CpuMemory[1] = memory;
        cartridge.CpuMemory[1].noBytes[0] = 0x1000;
        cartridge.CpuMemory[1].Bank[0] = new UCHAR [ 0x1000 ];
        memcpy ( cartridge.CpuMemory[1].Bank[0], cartridge.CpuMemory[0].Bank[0] + 0x1000, 0x1000 );
        sprintf ( name, "%stigpl.%s", base, ext );
        ReadHex ( name, cartridge );
    }
    if (( len >= 8 ) && (  strnicmp ( fileName + len - 8, "994a.", 5 ) == 0 )) {
        if ( len > 8 ) {
            base = fileName;
            fileName [ len - 8 ] = '\0';
	}
        sprintf ( name, "%s994arom.%s", base, ext );
        ReadHex ( name, cartridge );
        // Move GROM 3 to ROMs 0+1
        sMemoryRegion &memory = cartridge.GromMemory[3];
        cartridge.CpuMemory[0] = memory;
        cartridge.CpuMemory[0].noBytes[0] = 0x1000;
        cartridge.CpuMemory[1] = memory;
        cartridge.CpuMemory[1].noBytes[0] = 0x1000;
        cartridge.CpuMemory[1].Bank[0] = new UCHAR [ 0x1000 ];
        memcpy ( cartridge.CpuMemory[1].Bank[0], cartridge.CpuMemory[0].Bank[0] + 0x1000, 0x1000 );
        sprintf ( name, "%s994agrom.%s", base, ext );
        ReadHex ( name, cartridge );
    }

    // Move GROMs 3+4+5 to GROMs 0+1+2 and
    memcpy ( &cartridge.GromMemory[0], &cartridge.GromMemory[3], 3 * sizeof ( sMemoryRegion ));
    sMemoryRegion &memory = cartridge.GromMemory[6];
    cartridge.GromMemory[0].noBytes[0] = 0x1800;
    cartridge.GromMemory[1].noBytes[0] = 0x1800;
    cartridge.GromMemory[2].noBytes[0] = 0x1800;
    cartridge.GromMemory[3] = memory;
    cartridge.GromMemory[4] = memory;
    cartridge.GromMemory[5] = memory;

    // Create banks for 32K Memory Expansion and scratchpad RAM
    int ramBanks [] = { 0x02, 0x03, 8, 9, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F };

    for ( unsigned i = 0; i < SIZE ( ramBanks ); i++ ) {
        cartridge.CpuMemory[ramBanks[i]].Type       = _RAM_;
        cartridge.CpuMemory[ramBanks[i]].noBanks    = 1;
        cartridge.CpuMemory[ramBanks[i]].noBytes[0] = 0x1000;
    }

    // Set the name and save it to the expected filename
    cartridge.m_Title = strdup ( "TI-99/4A Console" );
    cartridge.SaveImage ( "TI-994A.ctg" );
}

//----------------------------------------------------------------------------
//
//    Gram Kracker: Loader/Saver Order and Header bytes
//
//    Byte        Description
//     00         More to load flag
//                  FF = More to load
//                  80 = Load UTIL Option next
//                  00 = Last file to load
//     01         What Gram Chip or Ram Bank
//                  01 = Grom/Gram 0 g0000
//                  02 = Grom/Gram 1 g2000
//                  03 = Grom/Gram 2 g4000
//                  04 = Grom/Gram 3 g6000
//                  05 = Grom/Gram 4 g8000
//                  06 = Grom/Gram 5 gA000
//                  07 = Grom/Gram 6 gC000
//                  08 = Grom/Gram 7 gE000
//                  09 = Rom/Ram Bank 1 c6000
//                  0A = Rom/Ram Bank 2 c6000
//            00 or FF = Program Image - load to Memory Expansion
//     02         0000 = Number of bytes to load
//     04         0000 = Address to start loading at
//
//----------------------------------------------------------------------------

void ReadGK ( char *baseFilename, cCartridge &cart )
{
    char filename [ 80 ];
    int more = 0xFF;

    for ( int i = 0; more == 0xFF; i++ ) {

        unsigned dummy = 0, bank = 0, length = 0, address= 0;

        sprintf ( filename, i ? "%s%d" : "%s", baseFilename, i );
        FILE *inFile = fopen ( filename, "rb" );
        if ( inFile == NULL ) {
            fprintf ( stderr, "Unable to open file %s\n", filename );
            return;
        }

        fread ( &dummy, 2, 1, inFile );
        if ( dummy != 0x8080 ) {
            fprintf ( stderr, "File %s is not a Gram Kracker file\n", filename );
            return;
        }

        fread ( &more, 1, 1, inFile );
        fread ( &bank, 1, 1, inFile );
        fread ( &length, 2, 1, inFile );
        fread ( &address, 2, 1, inFile );

        length = ( length >> 8 ) | (( length & 0xFF ) << 8 );
        address = ( address >> 8 ) | (( address & 0xFF ) << 8 );

        UCHAR *buffer = new UCHAR [ length ];
        fread ( buffer, length, 1, inFile );

        sMemoryRegion *region;

        if ( bank < 9 ) {
            region = &cart.GromMemory [ address >> 13 ];
            region->noBanks = 1;
            region->Bank[0] = buffer;
            region->noBytes[0] = length;
        } else if ( bank == 9 ) {
            region = &cart.CpuMemory [ address >> 12 ];
            if ( region->noBanks == 0 ) region->noBanks = 1;
            region->Bank[0] = buffer;
            region->noBytes[0] = length / 2;
            region++;
            UCHAR *newBuffer = new UCHAR [ length / 2 ];
            memcpy ( newBuffer, &buffer [ length / 2 ], length / 2 );
            if ( region->noBanks == 0 ) region->noBanks = 1;
            region->Bank[0] = newBuffer;
            region->noBytes[0] = length / 2;
        } else {
            region = &cart.CpuMemory [ address >> 12 ];
            region->noBanks = 2;
            region->Bank[1] = buffer;
            region->noBytes[1] = length / 2;
            region++;
            UCHAR *newBuffer = new UCHAR [ length / 2 ];
            memcpy ( newBuffer, &buffer [ length / 2 ], length / 2 );
            region->noBanks = 2;
            region->Bank[1] = newBuffer;
            region->noBytes[1] = length / 2;
        }

        fclose ( inFile );
    }
}

bool isCartridge ( char *fileName )
{
    int len = strlen ( fileName );
    return ( stricmp ( &fileName [len-4], ".ctg" ) == 0 ) ? true : false;
}

bool isHexFile ( char *fileName )
{
    int len = strlen ( fileName );
    if ( stricmp ( &fileName [len-4], ".hex" ) == 0 ) return true;
    return ( stricmp ( &fileName [len-4], ".bin" ) == 0 ) ? true : false;
}

bool isRomFile ( char *fileName )
{
    if ( isHexFile ( fileName ) == false ) return false;
    int len = strlen ( fileName );
    if (( len >= 6 ) && ( strnicmp ( fileName + len - 6, "ti.", 3 ) == 0 )) return true;
    if (( len >= 8 ) && ( strnicmp ( fileName + len - 8, "994a.", 5 ) == 0 )) return true;
    return false;
}

bool isListing ( char *fileName )
{
    int len = strlen ( fileName );
    if ( stricmp ( &fileName [len-4], ".lst" ) == 0 ) return true;
    if ( stricmp ( &fileName [len-4], ".dat" ) == 0 ) return true;
    return false;
}

eFileType WhichType ( char *fileName )
{
    if ( isCartridge ( fileName )) return CTG;
    if ( isRomFile ( fileName )) return ROM;
    if ( isHexFile ( fileName )) return HEX;
    if ( isListing ( fileName )) return LST;
    return GK;
}

int main ( int argc, char *argv[] )
{
    printf ( "Listing/.bin To TI-SIMULATOR .ctg file converter\n" );

    if ( argc == 1 ) {
        fprintf ( stderr, "\nSyntax: convert <input filename> [\"Cartridge Name\"]\n" );
        return -1;
    }
    cCartridge cartridge ( NULL );

    switch ( WhichType ( argv[1] )) {
        case CTG :
            printf ( "\nConverting .ctg file to compressed format\n" );
            cartridge.LoadImage ( argv[1] );
            cartridge.SaveImage ( argv[1] );
            return 0;
        case ROM :
            ReadConsole ( argv[1], cartridge );
            return 0;
        case HEX :
            ReadHex ( argv[1], cartridge );
            break;
        case LST :
            ReadFile ( argv[1], cartridge );
            break;
        case GK :
            ReadGK ( argv[1], cartridge );
            break;
        default :
            fprintf ( stderr, "Unable to determine the file format for %s\n", argv[1] );
            return -1;
    }

    char filename [256];
    sprintf ( filename,  ( argc == 3 ) ? argv[2] : "convert.ctg" );
    if ( strstr ( filename, ".ctg" ) == NULL ) strcat ( filename, ".ctg" );

    printf ( "\nModule Summary:\n" );

    printf ( "  GROMS: " );
    int groms = 0;
    for ( unsigned i = 0; i < SIZE ( cartridge.GromMemory ); i++ ) {
        if ( cartridge.GromMemory[i].noBytes[0] > 0 ) printf ( "%d ", i, groms++ );
    }
    printf ( groms ? "\n" : "None\n" );

    if ( cartridge.CpuMemory [0].noBytes[0] > 0 ) printf ( "  Operating System ROM\n" );

    for ( unsigned i = 2; i < SIZE ( cartridge.CpuMemory ); i++ ) {
        if ( cartridge.CpuMemory[i].noBytes[0] > 0 ) {
            int banks = cartridge.CpuMemory[i].noBanks;
            int type  = cartridge.CpuMemory[i].Type;
            printf ( "  %d bank%s of %s at %04X\n", banks, ( banks > 1 ) ? "s" : "", ( type == _ROM_ ) ? "ROM" : "RAM" , i * 0x1000 );
        }
    }

    printf ( "\n" );

    if ( argc == 3 ) {
        cartridge.m_Title = strdup ( argv[2] );
    } else {
        FindName ( cartridge );
    }
    CheckGrom ( cartridge );
    cartridge.SaveImage ( filename );

    return 0;
}
